"""Alpha shape   
"""

import exceptions
from Numeric import array, Float, divide_safe, sqrt, product
from load_mesh.loadASCII import load_points_file, export_boundary_file
import random

class PointError(exceptions.Exception): pass
class FlagError(exceptions.Exception): pass

OUTPUT_FILE_TITLE = "# The alpha shape boundary defined by point index pairs of edges"
INF = pow(10,9)

def alpha_shape_via_files(point_file, boundary_file, alpha= None):
    
    point_dict = load_points_file(point_file)
    points = point_dict['pointlist']
    #title_string = point_dict['title']
    
    alpha = Alpha_Shape(points)
    alpha.write_boundary(boundary_file)

class Alpha_Shape:

    def __init__(self, points, alpha = None):

        """ 
        Inputs:
        
          points: List of coordinate pairs [[x1, y1],[x2, y2]..] 

          alpha: alpha shape parameter
          
        """
        self._set_points(points)
        self._alpha_shape_algorithm(alpha)

    def _set_points(self, points):
        """
        """
        # print "setting points"
        if len (points) <= 2:
            raise PointError, "Too few points to find an alpha shape"
        if len(points)==3:
            #check not in a straight line
            x01 = points[0][0] - points[1][0]
            y01 = points[0][1] - points[1][1]
            x12 = points[1][0] - points[2][0]
            y12 = points[1][1] - points[2][1]
            crossprod = x01*y12 - x12*y01
            if crossprod==0:
                raise PointError, "Three points on a straight line"
        
        #Convert input to Numeric arrays
        self.points = array(points).astype(Float)

    
    def write_boundary(self,file_name):
        """
        Write the boundary to a file
        """
        #print " this info will be in the file" 
        export_boundary_file(file_name, self.get_boundary(),
                             OUTPUT_FILE_TITLE, delimiter = ',')
    
    def get_boundary(self):
        """
        """
        return self.boundary

    def set_boundary_type(self,raw_boundary=True,
                          remove_holes=False,
                          smooth_indents=False,
                          expand_pinch=False,
                          boundary_points_fraction=0.2):
        """
        Use the flags to set constraints on the boundary
        raw_boundary   Return raw boundary i.e. the regular edges
        remove_holes   filter to remove small holes
        smooth_indents   remove sharp indents in boundary
        expand_pinch   plus test for pinch-off
                       i.e. boundary vertex with more than two edges.
        """

        if raw_boundary:
            reg_edge = self.get_regular_edges(self.alpha)
            self.boundary = [self.edge[k] for k in reg_edge]
            self._init_boundary_triangles()
        if remove_holes:
            #remove holes
            self.boundary = self._remove_holes(boundary_points_fraction)
        if smooth_indents:
            #remove sharp indents
            self.boundary = self._smooth_indents()
        if expand_pinch:
            #deal with pinch-off
            self.boundary = self._expand_pinch()
        

    def get_delaunay(self):
        """
        """
        return self.deltri

    def get_optimum_alpha(self):
        """
        """
        return self.optimum_alpha

    def get_alpha(self):
        """
        Return current alpha value
        """
        return self.alpha

    def set_alpha(self,alpha):
        """
        Set alpha and update alpha-boundary. 
        """
        self.alpha = alpha
        reg_edge = self.get_regular_edges(alpha)    
        self.boundary = [self.edge[k] for k in reg_edge]
        self._init_boundary_triangles()
    
            
    def _alpha_shape_algorithm(self,alpha):
        
        # A python style guide suggests using _[function name] to
        # specify internal functions
        # - this is the first time I've used it though - DSG

        #print "starting alpha shape algorithm" 

        self.alpha = alpha

        # Build Delaunay triangulation
        import triang
        points = []
        seglist = []
        holelist = []
        regionlist = []
        pointattlist = []
        segattlist = []
        trilist = []

        points = [(pt[0], pt[1]) for pt in self.points]
        pointattlist = [ [] for i in range(len(points)) ] 
        mode = "Qzcn"
        #print "computing delaunay triangulation ... \n" 
        tridata = triang.genMesh(points,seglist,holelist,regionlist,pointattlist,segattlist,trilist,mode)
        #print tridata
        #print "point attribute list: ", tridata['generatedpointattributelist'],"\n"
        #print "hull segments: ", tridata['generatedsegmentlist'], "\n"
        self.deltri = tridata['generatedtrianglelist']
        self.deltrinbr = tridata['generatedtriangleneighborlist']
        self.hulledges = tridata['generatedsegmentlist']                              

        #print "Number of delaunay triangles: ", len(self.deltri), "\n"
        #print "deltrinbrs: ", self.deltrinbr, "\n"

        # Build Alpha table
        # print "Building alpha table ... \n" 
        self._tri_circumradius()
        # print "Largest circumradius ", max(self.triradius)
        self._edge_intervals()
        self._vertex_intervals()

        if alpha==None:
            # Find optimum alpha
            # Ken Clarkson's hull program uses smallest alpha so that
            # every vertex is non-singular so... 
            self.optimum_alpha = max([ iv[0] for iv in self.vertexinterval if iv!=[] ])
            # print "optimum alpha ", self.optimum_alpha
            self.alpha = self.optimum_alpha
            #alpha_tri = self.get_alpha_triangles(self.optimum_alpha)
            #print "alpha shape triangles ", alpha_tri

            reg_edge = self.get_regular_edges(self.optimum_alpha)
            #print "alpha boundary edges", reg_edge
        else:
            reg_edge = self.get_regular_edges(alpha)
    
        self.boundary = [self.edge[k] for k in reg_edge]
        #print "alpha boundary edges ", self.boundary
        self._init_boundary_triangles()
            
        return

    def _tri_circumradius(self):

        # compute circumradii of the delaunay triangles
        x = self.points[:,0]
        y = self.points[:,1]
        ind1 = [self.deltri[j][0] for j in range(len(self.deltri))]
        ind2 = [self.deltri[j][1] for j in range(len(self.deltri))]
        ind3 = [self.deltri[j][2] for j in range(len(self.deltri))]

        x1 = array([ x[j] for j in ind1 ])
        y1 = array([ y[j] for j in ind1 ])
        x2 = array([ x[j] for j in ind2 ])
        y2 = array([ y[j] for j in ind2 ])
        x3 = array([ x[j] for j in ind3 ])
        y3 = array([ y[j] for j in ind3 ])

        x21 = x2-x1
        x31 = x3-x1
        y21 = y2-y1
        y31 = y3-y1

        dist21 = x21*x21 + y21*y21
        dist31 = x31*x31 + y31*y31

        denom = x21*y31 - x31*y21
        #print "denom = ", denom
        delta = 0.00000001
        random.seed()
        zeroind = [k for k in range(len(denom)) if denom[k]==0 ]
        while zeroind!=[]:
            print "Warning: degenerate triangles found in alpha_shape.py, results may be inaccurate."
            for d in zeroind:
                x1[d] = x1[d]+delta*(random.random()-0.5)
                x2[d] = x2[d]+delta*(random.random()-0.5)
                x3[d] = x3[d]+delta*(random.random()-0.5)
                y1[d] = y1[d]+delta*(random.random()-0.5)
                y2[d] = y2[d]+delta*(random.random()-0.5)
                y3[d] = y3[d]+delta*(random.random()-0.5)
            x21 = x2-x1
            x31 = x3-x1
            y21 = y2-y1
            y31 = y3-y1
            dist21 = x21*x21 + y21*y21
            dist31 = x31*x31 + y31*y31
            denom = x21*y31 - x31*y21
            zeroind = [k for k in range(len(denom)) if denom[k]==0 ]

        # dx/2, dy/2 give circumcenter relative to x1,y1. 
        #dx = (y31*dist21 - y21*dist31)/denom
        #dy = (x21*dist31 - x31*dist21)/denom
        # from Numeric import divide_safe
        try:
            dx = divide_safe(y31*dist21 - y21*dist31,denom)
            dy = divide_safe(x21*dist31 - x31*dist21,denom)
        except ZeroDivisionError:
            print "There are serious problems with the delaunay triangles"
            raise
 
            
        self.triradius = 0.5*sqrt(dx*dx + dy*dy)
        #print "triangle radii", self.triradius
        return



    def _edge_intervals(self):
        # for each edge, find triples
        # (length/2, min_adj_triradius, max_adj_triradius) if unattached
        # (min_adj_triradius, min_adj_triradius, max_adj_triradius) if attached.

        # It should be possible to rewrite this routine in an array-friendly form
        # like _tri_circumradius()  if we need to speed things up. 

        edges = []
        edgenbrs = []
        edgeinterval = []
        for t in range(len(self.deltri)):
            tri = self.deltri[t]
            trinbr = self.deltrinbr[t]
            dx = array([self.points[tri[(i+1)%3],0] - self.points[tri[(i+2)%3],0] for i in [0,1,2]])
            dy = array([self.points[tri[(i+1)%3],1] - self.points[tri[(i+2)%3],1] for i in [0,1,2]])
            elen = sqrt(dx*dx+dy*dy)
            #angle = array([(-dx[(i+1)%3]*dx[(i+2)%3]-dy[(i+1)%3]*dy[(i+2)%3])/(elen[(i+1)%3]*elen[(i+2)%3]) for i in [0,1,2]])
            #angle = arccos(angle)
            # really only need sign - not angle value:
            anglesign = array([(-dx[(i+1)%3]*dx[(i+2)%3]-dy[(i+1)%3]*dy[(i+2)%3]) for i in [0,1,2]])
            
            #print "dx ",dx,"\n"
            #print "dy ",dy,"\n"
            #print "edge lengths of triangle ",t,"\t",elen,"\n"
            #print "angles ",angle,"\n"
                          
            for i in [0,1,2]:
                j = (i+1)%3
                k = (i+2)%3
                if trinbr[i]==-1:
                    edges.append((tri[j], tri[k]))
                    edgenbrs.append((t, -1))
                    edgeinterval.append([0.5*elen[i], self.triradius[t], INF])
                elif (tri[j]<tri[k]):
                    edges.append((tri[j], tri[k]))
                    edgenbrs.append((t, trinbr[i]))
                    edgeinterval.append([0.5*elen[i],\
                                         min(self.triradius[t],self.triradius[trinbr[i]]),\
                                             max(self.triradius[t],self.triradius[trinbr[i]]) ])
                else:
                    continue
                #if angle[i]>pi/2:
                if anglesign[i] < 0: 
                    edgeinterval[-1][0] = edgeinterval[-1][1]
                    
        self.edge = edges
        self.edgenbr = edgenbrs
        self.edgeinterval = edgeinterval
        #print "edges: ",edges, "\n"
        #print "edge nbrs:", edgenbrs ,"\n"
        #print "edge intervals: ",edgeinterval , "\n" 

    def _vertex_intervals(self):
        # for each vertex find pairs
        # (min_adj_triradius, max_adj_triradius)
        nv = len(self.points)
        vertexnbrs = [ [] for i in range(nv)]
        vertexinterval = [ [] for i in range(nv)] 
        for t in range(len(self.deltri)):
            for j in self.deltri[t]:
                vertexnbrs[j].append(t)
        for h in range(len(self.hulledges)):
            for j in self.hulledges[h]:
                vertexnbrs[j].append(-1)
        
        for i in range(nv):
            radii = [ self.triradius[t] for t in vertexnbrs[i] if t>=0 ]
            try:
                vertexinterval[i] = [min(radii), max(radii)]
                if vertexnbrs[i][-1]==-1:
                    vertexinterval[i][1]=INF
            except ValueError:
                # print "Vertex %i is isolated?"%i
                # print "coords: ",self.points[i]
                # print "Vertex nbrs: ", vertexnbrs[i]
                # print "nbr radii: ",radii
                # vertexinterval[i] = [0,INF]
                pass

        self.vertexnbr = vertexnbrs
        self.vertexinterval = vertexinterval
        #print "vertex nbrs ", vertexnbrs, "\n"
        #print "vertex intervals ",vertexinterval, "\n"
        
    def get_alpha_triangles(self,alpha):
        """
        Given an alpha value,
        return indices of triangles in the alpha-shape
        """
        def tri_rad_lta(k):
            return self.triradius[k]<=alpha

        return filter(tri_rad_lta, range(len(self.triradius)))

    def get_regular_edges(self,alpha):
        """
        Given an alpha value,
        return the indices of edges on the boundary of the alpha-shape
        """
        def reg_edge(k):
            return self.edgeinterval[k][1]<=alpha and self.edgeinterval[k][2]>alpha

        return filter(reg_edge, range(len(self.edgeinterval)))

    def get_exposed_vertices(self,alpha):
        """
        Given an alpha value,
        return the vertices on the boundary of the alpha-shape
        """
        def exp_vert(k):
            return self.vertexinterval[k][0]<=alpha and self.vertexinterval[k][1]>alpha

        return filter(exp_vert, range(len(self.vertexinterval)))        

    def _vertices_from_edges(self,elist):
        """
        Returns the list of unique vertex labels from edges in elist
        """

        v1 = [elist[k][0] for k in range(len(elist))]
        v2 = [elist[k][1] for k in range(len(elist))]
        v = v1+v2
        v.sort()
        vertices = [v[k] for k in range(len(v)) if v[k]!=v[k-1] ]
        return vertices

    def _init_boundary_triangles(self):
        """
        Creates the initial list of triangle indices
        exterior to and touching the boundary of the alpha shape
        """
        def tri_rad_gta(k):
            return self.triradius[k]>self.alpha

        extrind = filter(tri_rad_gta, range(len(self.triradius)))

        bv = self._vertices_from_edges(self.boundary)
        
        btri = []
        for et in extrind:
            v0 = self.deltri[et][0]
            v1 = self.deltri[et][1]
            v2 = self.deltri[et][2]
            if v0 in bv or v1 in bv or v2 in bv:
                btri.append(et)

        self.boundarytriangle = btri
  
        #print "exterior triangles: ", extrind

       
    def _remove_holes(self,small):
        """
        Given the edges in bdry, find the largest exterior components.
        """

        #print "running _remove_holes \n"

        bdry = self.boundary
        
        def findroot(i):
            if vptr[i] < 0:
                return i
            k = findroot(vptr[i])
            vptr[i] = k
            return k



        # get a list of unique vertex labels:
        verts = self._vertices_from_edges(bdry)
        #print "verts ", verts, "\n"

        #initialise vptr and eptr to negative number outside range 
        EMPTY = -max(verts)-len(verts)
        vptr = [EMPTY for k in range(len(verts))]
        #print "vptr init: ", vptr, "\n"

        #add edges and maintain union tree
        for i in range(len(bdry)):
            #print "edge ",i,"\t",bdry[i] 
            vl = verts.index(bdry[i][0])
            vr = verts.index(bdry[i][1])
            #rvl = vl, rvr = vr
            rvl = findroot(vl)
            rvr = findroot(vr)
            #print "roots:  ",rvl, rvr
            if not(rvl==rvr):
                if (vptr[vl]==EMPTY):
                    if (vptr[vr]==EMPTY):
                        vptr[vl] = -2
                        vptr[vr] = vl
                    else:
                        vptr[vl] = rvr
                        vptr[rvr] = vptr[rvr]-1
                else:
                    if (vptr[vr]==EMPTY):
                        vptr[vr] = rvl
                        vptr[rvl] = vptr[rvl]-1
                    else:
                        if vptr[rvl] > vptr[rvr]:
                            vptr[rvr] = vptr[rvr] + vptr[rvl]
                            vptr[rvl] = rvr
                            vptr[vl] = rvr
                        else:
                            vptr[rvl] = vptr[rvl] + vptr[rvr]
                            vptr[rvr] = rvl
                            vptr[vr] = rvl 
                #print "vptr: ", vptr, "\n"

        # end edge loop

        # print "vertex component tree: ", [-v for v in vptr if v<0], "\n"
        # print "largest component has: ", -min(vptr), " points. \n"
        # discard the edges in the little components
        # (i.e. those components with less than 'small' fraction of bdry points)
        cutoff = round(small*len(verts))
        # print "cutoff component size is ", cutoff, "\n"
        largest_component = -min(vptr)
        if cutoff > largest_component:
            cutoff = round((1-small)*largest_component)
        
        littleind = [k for k in range(len(vptr)) if (vptr[k]<0 and vptr[k]>-cutoff)] 
        if littleind:
            littlecomp = [k for k in range(len(vptr)) if findroot(k) in littleind]
            vdiscard = [verts[k] for k in littlecomp] 
            newbdry = [e for e in bdry if not((e[0] in vdiscard) and (e[1] in vdiscard))]
            # remove boundary triangles that touch discarded components
            newbt = []
            for bt in self.boundarytriangle:
                v0 = self.deltri[bt][0]
                v1 = self.deltri[bt][1]
                v2 = self.deltri[bt][2]
                if not (v0 in vdiscard or v1 in vdiscard or v2 in vdiscard):
                    newbt.append(bt)
            self.boundarytriangle = newbt
        else:
            newbdry = bdry
    
        return newbdry


    def _smooth_indents(self):
        """
        Given edges in bdry, test for acute-angle indents and remove them.
        """
        
        #print "running _smooth_indents \n"

        bdry = self.boundary
        bdrytri = self.boundarytriangle
        verts = self._vertices_from_edges(bdry)
        
        # find boundary triangles that have two edges in bdry
        b2etri = []
        for ind in bdrytri:
            bect = 0
            for j in [0,1,2]:
                eda = (self.deltri[ind][(j+1)%3], self.deltri[ind][(j+2)%3])
                edb = (self.deltri[ind][(j+2)%3], self.deltri[ind][(j+1)%3])               
                if eda in bdry or edb in bdry:
                    bect +=1
            if bect==2:
                b2etri.append(ind)

        # test the bdrytri triangles for acute angles 
        acutetri = []
        for tind in b2etri:
            tri = self.deltri[tind]
            dx = array([self.points[tri[(i+1)%3],0] - self.points[tri[(i+2)%3],0] for i in [0,1,2]])
            dy = array([self.points[tri[(i+1)%3],1] - self.points[tri[(i+2)%3],1] for i in [0,1,2]])
            anglesign = array([(-dx[(i+1)%3]*dx[(i+2)%3]-dy[(i+1)%3]*dy[(i+2)%3]) for i in [0,1,2]])
            if product(anglesign) > 0:
                acutetri.append(tind)

        #print "acute boundary triangles ", acutetri

        # adjust the bdry edges and triangles by adding in the acutetri triangles
        for pind in acutetri:
            bdrytri.remove(pind) 
            tri = self.deltri[pind]
            for i in [0,1,2]:
                bdry.append((tri[(i+1)%3], tri[(i+2)%3]))

        self.boundarytriangle = bdrytri

        newbdry = []
        for ed in bdry:
            numed = bdry.count(ed)+bdry.count((ed[1],ed[0]))
            if numed%2 == 1:
                newbdry.append(ed)
        
        #print "new boundary ", newbdry
        return newbdry

    def _expand_pinch(self):
        """
        Given edges in bdry, test for vertices with more than 2 incident edges.
        Expand by adding back in associated triangles... 
        """
        #print "running _expand_pinch \n"

        bdry = self.boundary
        bdrytri = self.boundarytriangle
        
        v1 = [bdry[k][0] for k in range(len(bdry))]
        v2 = [bdry[k][1] for k in range(len(bdry))]
        v = v1+v2
        v.sort()
        probv = [v[k] for k in range(len(v)) if (v[k]!=v[k-1] and v.count(v[k])>2) ]
        #print "problem vertices: ", probv  

        # find boundary triangles that have at least one vertex in probv
        probtri = []
        for ind in bdrytri:
            v0 = self.deltri[ind][0]
            v1 = self.deltri[ind][1]
            v2 = self.deltri[ind][2]
            if v0 in probv or v1 in probv or v2 in probv:
                probtri.append(ind)

        #print "problem boundary triangle indices ", probtri

        # "add in" the problem triangles
        for pind in probtri:
            bdrytri.remove(pind) 
            tri = self.deltri[pind]
            for i in [0,1,2]:
                bdry.append((tri[(i+1)%3], tri[(i+2)%3]))

        self.boundarytriangle = bdrytri

        newbdry = []
        for ed in bdry:
            numed = bdry.count(ed)+bdry.count((ed[1],ed[0]))
            if numed%2 == 1:
                newbdry.append(ed)
        
        #print "new boundary ", newbdry
        return newbdry


#-------------------------------------------------------------
if __name__ == "__main__":
    """
    Load in a data point file.
    Determine the alpha shape boundary
    Save the boundary to a file.

    usage: alpha_shape.py point_file.xya boundary_file.bnd [alpha]
    
    The alpha value is optional.
    """
    
    import os, sys
    usage = "usage: %s point_file.xya boundary_file.bnd [alpha]"%os.path.basename(sys.argv[0])
    # I made up the .bnd affix. Other ideas welcome. -DSG 
    if len(sys.argv) < 3:
        print usage
    else:
        point_file = sys.argv[1]
        boundary_file = sys.argv[2]
        if len(sys.argv) > 4:
            alpha = sys.argv[3]
        else:
            alpha = None

        #print "about to call alpha shape routine \n"
        alpha_shape_via_files(point_file, boundary_file, alpha)
        
